from .work_unit import Priority, WorkUnit
from .task import Task


class Pool(WorkUnit):
    """A group of tasks. Can be embedded into another pool.

    A pool doesn't represent a task by itself. It's only
    a container for tasks.

    Args:
        custom_fields (Optional[Dict[str, Any]):
            Custom fields for the pool. Not used in plan calculations.
            Can be accessed by keys:
            >>> pool['field_name']
        parent (Optional[Pool]):
            Parent pool.
        priority (Optional[Priority]):
            The pool's priority. Defaults to "normal".
        children (Optional[List[Task | Pool]]):
            List of child tasks and pools, embedded into this pool.
        id_ (Optional[str]):
            Unique identifier for the pool. Generated if not provided.
        name (Optional[str]):
            Name of the pool.
    """

    def __init__(
        self,
        custom_fields=None,
        parent=None,
        priority=Priority("normal"),
        children=[],
        id_=None,
        name=None,
    ):
        self.children = {}
        self._indexate_pool(children)

        super().__init__(custom_fields, parent, priority, id_, name)

    def _indexate_pool(self, pool):
        """Indexate the given pool's children into a dict for fast access.

        Args:
            pool (Pool):
                A pool to indexate.
        """
        for c in pool:
            if c.id in self.children:
                raise ValueError(
                    (
                        f"Child '{c.name}' with id '{c.id}' already"
                        " exists in the pool {self.name}."
                    )
                )

            if isinstance(c, Pool):
                self._indexate_pool(c)

            if not c.parent:
                c.parent = self

            self.children[c.id] = c

    def __iter__(self):
        """Iterate over the pool's children.

        Returns:
            Iterator[Task | Pool]:
                An iterator over this pool's children.
        """
        return iter(filter(lambda c: c.parent == self, self.children.values()))

    def __len__(self):
        """This pool's length.

        Returns:
            int: The number of children in this pool.
        """
        return len(self.children)

    def __repr__(self):
        """String representation of the pool.

        Returns:
            str:
        """
        return f"Pool(id={self.id}, name={self.name})"

    @property
    def dry(self):
        """A minified version of the pool.

        Used during plans calculation to reduce the amount of data
        to process.

        Returns:
            Dict[str, Dict[str, Any]]:
                A dictionary with the pool's children dry versions.
        """
        todo = list(
            filter(
                lambda c: not c.is_solved and isinstance(c, Task),
                self.children.values(),
            )
        )
        todo.sort(key=lambda t: t.cost)

        dry = {t.id: t.dry for t in todo}
        return dry

    @property
    def done(self):
        """Tasks from this pool that are done.

        Returns:
            Iterator[Task]:
        """
        for t in filter(
            lambda c: c.is_solved and c.parent == self, self.children.values()
        ):
            yield t

    @property
    def is_solved(self):
        """This pool solved status.

        Returns:
            bool: True if all tasks in this pool are solved, False otherwise.
        """
        for c in self.children.values():
            if not c.is_solved:
                return False

        return True

    @property
    def todo(self):
        """Tasks from this pool that are not done.

        Returns:
            Iterator[Task]:
        """
        for t in filter(
            lambda c: not c.is_solved and c.parent == self,
            self.children.values(),
        ):
            yield t

    def add(self, child):
        """Add a task or pool into this pool.

        Args:
            child (Task | Pool):
                A task or pool to add into this pool.
        """
        if child.id in self.children:
            raise ValueError(
                (
                    f"Child with id '{child.id}' already exists "
                    "in the pool {self.id}."
                )
            )

        child.parent = self
        self.children[child.id] = child

        if isinstance(child, Pool):
            self._indexate_pool(child)

    def get(self, child_id):
        """Return a child from this pool by its id.

        Can return children of the embedded pools.

        Args:
            child_id (uuid.UUID | str):
                The id of the child to return.

        Returns:
            Task | Pool:
        """
        return self.children[child_id]

    def pop(self, child_id):
        """Remove a child from this pool by its id.

        Can remove children of the embedded pools.

        Args:
            child_id (uuid.UUID | str):
                The id of the child to remove.

        Returns:
            Task | Pool:
                The removed child.
        """
        child = self.children[child_id]
        child.parent._pop(child_id)

        child.parent = None
        return child

    def _pop(self, child_id):
        """
        Convenience method to recursively remove a child
        from this and the parent pools.

        Args:
            child_id (uuid.UUID | str):
                The id of the child to remove.
        """
        del self.children[child_id]
        if self.parent:
            self.parent._pop(child_id)



"""Energoton classes."""

from ..base import Id
from ..relation import Blocking
from ..planner import Plan


class Energoton(Id):
    """Base class for energotons.

    Walks through the given pool of tasks, building all possible plans,
    chooses the best ones and returns them to the user.

    Args:
        capacity (Union[int, List[int]]):
            The amount of energy the energoton has to solve tasks.
            If it's an `int`, on every recharge, energoton's capacity
            will be restored to this value. If it's a `list`, energoton's
            capacity will be taken from the list, cycle by cycle until
            the list is empty.
        id_ (Optional[Union[str, uuid.UUID]]):
            The ID of the energoton. Will be generated if not provided.
        name (Optional[str]):
            The name of the energoton.
    """

    def __init__(self, capacity, id_=None, name=None):
        self.name = name
        self._capacity = capacity
        self.energy_left = self.next_charge

        self.pool = None
        self._dry_pool = None

        self._max_plan_value = 0
        self._cur_plan_value = 0

        self.is_dtrm = isinstance(self, DeterministicEnergoton)

        super().__init__(id_)

    def __eq__(self, other):
        """Equality operator.

        Args:
            other (Energoton): The other energoton to compare with.

        Returns:
            bool: True if the energotons are equal, False otherwise.
        """
        return self.id == other.id

    @property
    def capacity(self):
        """The energoton's capacity.

        Returns:
            Union[int, List[int]]: The energoton's capacity.
        """
        return self._capacity

    @property
    def next_charge(self):
        """Get the next energoton's charge.

        Returns:
            int: The energoton's charge for the next cycle.
        """
        if isinstance(self._capacity, int):
            return self._capacity

        if isinstance(self._capacity, list):
            if len(self._capacity) == 0:
                return 0

            return self._capacity.pop(0)

    def recharge(self):
        """Recharge the energoton."""
        self.energy_left = self.next_charge

    def _commit_plan(self, plan, plans, tasks):
        """Commit the given plan.

        If the plan has a lower value than the current best plan,
        it'll be discarded from further considerations. If the plan
        has a higher value, all the currently stored plans will be
        discarded.

        Args:
            plan (Plan): The plan to commit.
            plans (List[Plan]): Currently considered plans.
            tasks (List[Dict]): The current pool state.
        """
        if plans:
            if self._cur_plan_value < plans[0].value:
                return

            sorted_plan = Plan(sorted(plan, key=sort_key))
            sorted_plan.commit(self._cur_plan_value, tasks)

            if sorted_plan.value > plans[0].value:
                plans.clear()
                self._max_plan_value = sorted_plan.value
            elif sorted_plan in plans:
                return
        else:
            sorted_plan = Plan(sorted(plan, key=sort_key))
            sorted_plan.commit(self._cur_plan_value, tasks)

        plans.append(sorted_plan)

    def _build_plans(self, task, plan, tasks, plans, cycle, ind=0):
        """Recursively build plans with depth-first search.

        Args:
            task (Optional[Dict]): The current task to work on.
            plan (Plan): The current plan being built.
            tasks (List[Dict]): The list of tasks to work on.
            plans (List[Plan]): The list of plans being built.
            cycle (int): The current cycle number.
            ind (int): The index of the current task in the list.
        """
        if task:
            # work on a task
            energy_spent = (
                task["left"]
                if task["left"] <= self.energy_left
                else self.energy_left
            )

            self.energy_left -= energy_spent
            task["left"] -= energy_spent

            or_task = self.pool.children[task["id"]]
            plan.append(
                {
                    "task": or_task,
                    "amount": energy_spent,
                    "assignee": self,
                    "cycle": cycle,
                }
            )
            if self.is_dtrm:
                self._cur_plan_value += or_task.priority.value
            else:
                self._cur_plan_value += or_task.priority.value * (
                    energy_spent / or_task.cost
                )

            if not task["left"]:
                del tasks[ind]

        if not self.energy_left:
            if self._cur_plan_value >= self._max_plan_value:
                self._commit_plan(plan, plans, tasks)
        else:
            # see if it's possible to continue working
            # on the tasks that left
            can_continue = False
            for i, t in enumerate(tasks):
                if self.energy_left < t["left"] and self.is_dtrm:
                    break

                if self.pool.children[t["id"]].relations:
                    if self._is_actual(t["id"]):
                        self._build_plans(t, plan, tasks, plans, cycle, i)
                        can_continue = True
                else:
                    self._build_plans(t, plan, tasks, plans, cycle, i)
                    can_continue = True

            if not can_continue:
                if self._cur_plan_value >= self._max_plan_value:
                    self._commit_plan(plan, plans, tasks)

        if task:
            # reverse all the changes to the pool one step back
            if not task["left"]:
                tasks.insert(ind, task)

            del plan[-1]

            self.energy_left += energy_spent
            task["left"] += energy_spent
            self._cur_plan_value -= or_task.priority.value

    def build_plans(self, dry_pool, cycle=1, plan=None):
        """The main method to build plans.

        Args:
            dry_pool (Dict): The pool of tasks to work on.
            cycle (int): The current cycle number.
            plan (Optional[Plan]):
                The current plan being built. Can include tasks from
                the previous work cycle.
        """
        self._dry_pool = dry_pool
        self._cur_plan_value = plan.value if plan else 0

        plans = []
        self._build_plans(
            task=None,
            plan=plan or Plan(),
            tasks=list(dry_pool.values()),
            plans=plans,
            cycle=cycle,
        )
        return plans

    def _is_actual(self, task_id):
        """Check the given task's relations.

        Args:
            task_id (str): The ID of the task to check.

        Returns:
            bool:
                False if the task is blocked by another task,
                or its alternative is already solved. True otherwise.
        """
        for rel in self.pool.children[task_id].relations.values():
            if isinstance(rel, Blocking):
                if rel.blocked.id == task_id:
                    if rel.blocker.id not in self._dry_pool:
                        continue

                    dry = self._dry_pool[rel.blocker.id]
                    if dry["left"]:
                        return False
            else:
                for alt in rel.alternatives:
                    dry = self._dry_pool[alt.id]
                    if dry["left"] == 0:
                        return False

        return True


class DeterministicEnergoton(Energoton):
    """Transactional type of worker.

    It solves a task, only if it has enough energy to finish it.
    """

    pass


class NonDeterministicEnergoton(Energoton):
    """Partial type of worker.

    It can on a task, even if it doesn't have enough energy to finish it.
    """

    pass


def sort_key(w):
    """Sort key for work logs in a plan.

    Args:
        w (Dict): The work log to sort.

    Returns:
        str: The ID of the task in the work log.
    """
    return w["task"].id



import abc

from energoton.base import Id

from ..relation import Alternative, Blocking


class Priority:
    """A work unit priority.

    Priority is used to determine the task value. The higher
    is the priority, the higher is the value, and the more
    chance that the energoton will choose this task for solving.

    Args:
        label (Literal["lowest", "low", "normal", "high", "highest", "must"]):
            Priority label.
    """

    exp_values = {
        "lowest": 1,
        "low": 2,
        "normal": 4,
        "high": 8,
        "highest": 16,
        "must": 1000000,
    }

    def __init__(self, label):
        self.label = label
        self.value = self.exp_values[label]

    def __repr__(self):
        return f"Priority('{self.label}')"


class WorkUnit(Id, metaclass=abc.ABCMeta):
    """Represents a single work unit to do. Can be a task or a pool.

    Args:
        custom_fields (Optional[Dict[str: Any]]):
            Custom fields for the work unit. Can be accessed by keys:
            >>> work_unit['field_name']
        parent (Optional[Pool]):
            Parent pool.
        priority (Optional[Priority]):
            Priority of the work unit. Defaults to "medium" priority.
        id_ (Optional[str | uuid.UUID]):
            Id of the work unit. Generated, if not provided explicitly.
        name (Optional[str]):
            Name of the work unit.
    """

    def __init__(
        self,
        custom_fields=None,
        parent=None,
        priority=Priority("normal"),
        id_=None,
        name=None,
    ):
        super().__init__(id_)

        self.custom_fields = custom_fields or {}
        self.name = name
        self.relations = {}
        self.priority = priority
        self.parent = parent

    @abc.abstractmethod
    def is_solved(self):
        """Check if the work unit is solved.

        Raises:
            NotImplementedError:
                Must be implemented for any work unit.
        """
        raise NotImplementedError(
            "is_solved method must be implemented for any work unit!"
        )

    def __delitem__(self, key):
        """A custom field deletion.

        Args:
            key (str): Field name.
        """
        del self.custom_fields[key]

    def __getitem__(self, key):
        """A custom field access.

        Args:
            key (str): Field name.

        Returns:
            Any: Custom field value.
        """
        return self.custom_fields[key]

    def __setitem__(self, key, value):
        """A custom field setting."

        Args:
            key (str): Field name.
            value (Any): Field value.
        """
        self.custom_fields[key] = value

    @property
    def blocked_by(self):
        """
        Return all relations that prevent this work
        unit from being done.

        Yields:
            work.relation.Blocking: Blocking relations.
        """
        for rel in self.relations.values():
            if (
                isinstance(rel, Blocking)
                and rel.blocked == self
                and not rel.blocker.is_solved
            ):
                yield rel

    @property
    def blocking(self):
        """
        Return all relations where this work unit
        prevents other units from being solved.

        Yields:
            work.relation.Blocking: Blocking relations.
        """
        for rel in self.relations.values():
            if (
                isinstance(rel, Blocking)
                and rel.blocker == self
                and not self.is_solved
            ):
                yield rel

    @property
    def is_actual(self):
        """Check if this unit still needs to be solved.

        Returns:
            bool:
                True if the unit is not solved and its
                alternatives are not solved as well.
        """
        for rel in self.relations.values():
            if isinstance(rel, Alternative) and rel.is_solved:
                return False

        return True

    @property
    def is_blocked(self):
        """Check if this unit is blocked by another unit.

        Returns:
           bool: True if the unit is blocked, False otherwise.
        """
        try:
            next(self.blocked_by)
            return True
        except StopIteration:
            return self.parent.is_blocked if self.parent else False



"""Task represents a single unit of work to be done."""

from .work_unit import Priority, WorkUnit


class Task(WorkUnit):
    """A piece of work to do.

    Args:
        cost (int):
            Energy cost of the task.
        custom_fields (Optional[Dict[str: Any]]):
            Custom fields for additional task info.
            Can be accessed by keys:
            >>> task['field_name']
        parent (Optional[Pool]):
            Parent pool.
        priority (Optional[Priority]):
            Priority of the task. Defaults to "medium" priority.
        id_ (Optional[str | uuid.UUID]):
            Id of the task. Generated, if not provided explicitly.
        name (Optional[str]):
            Name of the task.
    """

    def __init__(
        self,
        cost,
        custom_fields={},
        parent=None,
        priority=Priority("normal"),
        id_=None,
        name=None,
    ):
        self.work_done = []
        self.cost = cost

        super().__init__(custom_fields, parent, priority, id_, name)

    def __repr__(self):
        """Textual representation of the task.

        Returns:
            str: Textual representation of the task.
        """
        return (
            f"Task(id_='{self.id}', name='{self.name}', "
            f"cost={self.cost}, priority='{self.priority.label}')"
        )

    @property
    def dry(self):
        """Short form of the task, used during plan building.

        Returns:
            dict: Short form of the task.
        """
        return {
            "id": self.id,
            "cost": self.cost,
            "left": self.cost - self.spent,
            "priority": self.priority,
        }

    @property
    def spent(self):
        """Amount of energy spent on the task.

        Returns:
            int: Energy spent on the task.
        """
        return sum(w["amount"] for w in self.work_done)

    @property
    def is_solved(self):
        """Check if the task is solved.

        Returns:
            bool: True if the task is solved, False otherwise.
        """
        return self.spent == self.cost

    @property
    def todo(self):
        """Amount of energy left to be spent on the task.

        Returns:
            int: Energy left to be spent on the task.
        """
        return self.cost - self.spent



import unittest
from unittest import mock

from energoton import Task, Alternative, Blocking


class TestTask(unittest.TestCase):
    def test_status(self):
        task = Task(10, name="Task 1")
        self.assertFalse(task.is_solved)
        self.assertEqual(task.todo, 10)

        task.work_done.append(
            {"task": task, "amount": 10, "assignee": mock.Mock()}
        )
        self.assertTrue(task.is_solved)
        self.assertEqual(task.todo, 0)

        task.work_done = []
        task.work_done.append(
            {"task": task, "amount": 5, "assignee": mock.Mock()}
        )
        self.assertFalse(task.is_solved)
        self.assertEqual(task.todo, 5)

    def test_alternative_is_solved(self):
        t1 = Task(2, name="test-name1")
        t2 = Task(2, name="test-name2")
        t3 = Task(2, name="test-name3")

        rel = Alternative(t1, t2, t3)

        self.assertFalse(rel.is_solved)
        self.assertTrue(t1.is_actual)

        t1.work_done.append({"task": t1, "amount": 2, "assignee": mock.Mock()})
        self.assertTrue(rel.is_solved)
        self.assertFalse(t2.is_actual)
        self.assertFalse(t3.is_actual)

        t1.work_done = []
        t1.work_done.append({"task": t2, "amount": 2, "assignee": mock.Mock()})
        self.assertTrue(rel.is_solved)
        self.assertFalse(t1.is_actual)
        self.assertFalse(t3.is_actual)

    def test_part_done(self):
        task = Task(8, name="Task 1")

        work_done = {"task": task, "amount": 5, "assignee": mock.Mock()}
        task.work_done.append(work_done)

        self.assertEqual(work_done["amount"], 5)
        self.assertEqual(work_done["task"].spent, 5)
        self.assertEqual(work_done["task"].todo, 3)
        self.assertEqual(work_done["task"].name, "Task 1")

    def test_custom_fields(self):
        key1 = "key1"
        value1 = "value1"

        unit = Task(cost=3, name="test-name", custom_fields={key1: value1})

        self.assertEqual(unit[key1], value1)

        key2 = "key2"
        value2 = "value2"
        unit[key2] = value2

        self.assertEqual(unit[key2], value2)

        del unit[key2]

        with self.assertRaises(KeyError):
            unit[key2]

    def test_blocking_relationship(self):
        blocker = Task(cost=3, name="test-name1")
        blocked = Task(cost=3, name="test-name2")

        rel = Blocking(blocker, blocked)

        self.assertEqual(list(blocked.blocked_by), [rel])
        self.assertEqual(list(blocker.blocked_by), [])

        self.assertEqual(list(blocked.blocking), [])
        self.assertEqual(list(blocker.blocking), [rel])

    def test_is_blocked_by_parent(self):
        root = Task(cost=3, name="test-name4")
        parent = Task(cost=3, name="test-name3", parent=root)

        blocked = Task(cost=3, name="test-name2", parent=parent)
        blocker = Task(cost=3, name="test-name1")

        Blocking(blocker, parent)

        self.assertTrue(blocked.is_blocked)

        blocked.relations = {}
        parent.relations = {}
        self.assertFalse(blocked.is_blocked)

        Blocking(blocker, root)
        self.assertTrue(blocked.is_blocked)

    def test_drop_blocking(self):
        blocker = Task(cost=3, name="test-name1")
        blocked = Task(cost=3, name="test-name2")

        rel = Blocking(blocker, blocked)
        self.assertTrue(blocked.is_blocked)

        rel.drop()

        self.assertEqual(list(blocker.blocking), [])
        self.assertEqual(list(blocked.blocked_by), [])
        self.assertFalse(blocked.is_blocked)

    def test_alternative_relationship(self):
        alt1 = Task(cost=3, name="test-name1")
        alt2 = Task(cost=3, name="test-name2")
        alt3 = Task(cost=3, name="test-name3")

        rel = Alternative(alt1, alt2, alt3)

        self.assertEqual(list(alt1.relations.values())[0], rel)
        self.assertEqual(list(alt2.relations.values())[0], rel)
        self.assertEqual(list(alt3.relations.values())[0], rel)


class TestPartTask(unittest.TestCase):
    def test_part(self):
        task = Task(10, name="Task 1")

        done = {"task": task, "amount": 5, "assignee": mock.Mock()}
        task.work_done.append(done)

        self.assertEqual(done["amount"], 5)
        self.assertEqual(done["task"].spent, 5)
        self.assertEqual(done["task"].todo, 5)
        self.assertEqual(done["task"].name, "Task 1")



import unittest

from energoton.base import Id


class TestId(unittest.TestCase):
    def test_id_generated(self):
        unit1 = Id()
        unit2 = Id()

        self.assertNotEqual(unit1.id, unit2.id)

    def test_id_equality(self):
        unit1 = Id()
        unit2 = Id()

        self.assertNotEqual(unit1, unit2)

        unit2 = Id(unit1.id)
        self.assertEqual(unit1, unit2)



import unittest

from energoton import (
    Alternative,
    Blocking,
    Pool,
    Task,
    DeterministicEnergoton,
    NonDeterministicEnergoton,
)
from energoton.planner import Plan


class TestEnergoton(unittest.TestCase):
    def test_build_plans_deterministic(self):
        pool = Pool()
        t1 = Task(5, id_=1)
        t2 = Task(2, id_=2)
        t3 = Task(4, id_=3)
        t4 = Task(2, id_=4)
        t5 = Task(6, id_=5)

        pool.add(t1)
        pool.add(t2)
        pool.add(t3)
        pool.add(t4)
        pool.add(t5)

        e = DeterministicEnergoton(8)
        e.pool = pool
        plans = e.build_plans(pool.dry)

        p = Plan(
            [
                {"task": t2, "amount": 2, "assignee": e, "dry": (t2.id, 2)},
                {"task": t3, "amount": 4, "assignee": e, "dry": (t3.id, 4)},
                {"task": t4, "amount": 2, "assignee": e, "dry": (t4.id, 2)},
            ]
        )
        p.commit(0, ())

        self.assertEqual(plans, [p])

    def test_build_plans_non_deterministic(self):
        pool = Pool()
        t1 = Task(5, id_="1")
        t2 = Task(2, id_="2")
        t3 = Task(4, id_="3")

        pool.add(t1)
        pool.add(t2)
        pool.add(t3)

        e = NonDeterministicEnergoton(8)
        e.pool = pool
        plans = e.build_plans(pool.dry)

        p1 = Plan(
            [
                {"task": t1, "amount": 2, "assignee": e, "dry": (t1.id, 2)},
                {"task": t2, "amount": 2, "assignee": e, "dry": (t2.id, 2)},
                {"task": t3, "amount": 4, "assignee": e, "dry": (t3.id, 4)},
            ]
        )
        p1.commit(0, ())

        self.assertEqual(plans, [p1])

    def test_build_plans_blocked(self):
        pool = Pool()
        t1 = Task(5, id_="1")
        t2 = Task(2, id_="2")
        t3 = Task(4, id_="3")
        t4 = Task(2, id_="4")
        t5 = Task(6, id_="5")

        pool.add(t1)
        pool.add(t2)
        pool.add(t3)
        pool.add(t4)
        pool.add(t5)

        Blocking(t5, t3)

        e = DeterministicEnergoton(8)
        e.pool = pool
        plans = e.build_plans(pool.dry)

        p1 = Plan(
            [
                {"task": t2, "amount": 2, "assignee": e, "dry": (t2.id, 2)},
                {"task": t4, "amount": 2, "assignee": e, "dry": (t4.id, 2)},
            ]
        )
        p2 = Plan(
            [
                {"task": t1, "amount": 5, "assignee": e, "dry": (t1.id, 5)},
                {"task": t2, "amount": 2, "assignee": e, "dry": (t2.id, 2)},
            ]
        )
        p3 = Plan(
            [
                {"task": t2, "amount": 2, "assignee": e, "dry": (t2.id, 2)},
                {"task": t5, "amount": 6, "assignee": e, "dry": (t5.id, 6)},
            ]
        )
        p4 = Plan(
            [
                {"task": t1, "amount": 5, "assignee": e, "dry": (t1.id, 5)},
                {"task": t4, "amount": 2, "assignee": e, "dry": (t4.id, 2)},
            ]
        )
        p5 = Plan(
            [
                {"task": t4, "amount": 2, "assignee": e, "dry": (t4.id, 2)},
                {"task": t5, "amount": 6, "assignee": e, "dry": (t5.id, 6)},
            ]
        )
        p1.commit(0, ())
        p2.commit(0, ())
        p3.commit(0, ())
        p4.commit(0, ())
        p5.commit(0, ())

        self.assertEqual(
            plans,
            [p1, p2, p3, p4, p5],
        )

    def test_build_plans_alternative(self):
        pool = Pool(name="Pool")
        t1 = Task(5, id_="1")
        t2 = Task(2, id_="2")
        t3 = Task(4, id_="3")
        t4 = Task(2, id_="4")
        t5 = Task(6, id_="5")

        pool.add(t1)
        pool.add(t2)
        pool.add(t3)
        pool.add(t4)
        pool.add(t5)

        Alternative(t1, t2)

        e = DeterministicEnergoton(8)
        e.pool = pool
        plans = list(e.build_plans(pool.dry))

        p = Plan(
            [
                {"task": t2, "amount": 2, "assignee": e, "dry": (t2.id, 2)},
                {"task": t3, "amount": 4, "assignee": e, "dry": (t3.id, 4)},
                {"task": t4, "amount": 2, "assignee": e, "dry": (t4.id, 2)},
            ]
        )
        p.commit(0, ())

        self.assertEqual(plans, [p])

    def test_charges(self):
        charge = 5
        e = DeterministicEnergoton(charge)
        self.assertEqual(e.energy_left, charge)
        self.assertEqual(e.next_charge, charge)

        e = DeterministicEnergoton([5, 3, 2])
        self.assertEqual(e.energy_left, 5)
        e.recharge()

        self.assertEqual(e.energy_left, 3)
        e.recharge()

        self.assertEqual(e.energy_left, 2)
        e.recharge()

        self.assertEqual(e.energy_left, 0)



import unittest

from energoton.planner import Planner, Plan
from energoton import (
    Pool,
    Priority,
    Task,
    Blocking,
    DeterministicEnergoton,
    NonDeterministicEnergoton,
)


class TestPlanner(unittest.TestCase):
    def test_sort_by_value(self):
        pool = Pool()
        t1 = Task(
            5,
            id_=1,
            priority=Priority("high"),
        )
        t2 = Task(
            2,
            id_=2,
            priority=Priority("low"),
        )
        t3 = Task(
            4,
            id_=3,
            priority=Priority("normal"),
        )
        t4 = Task(
            2,
            id_=4,
            priority=Priority("highest"),
        )
        t5 = Task(
            6,
            id_=5,
            priority=Priority("lowest"),
        )

        pool.add(t1)
        pool.add(t2)
        pool.add(t3)
        pool.add(t4)
        pool.add(t5)

        e = DeterministicEnergoton(8, name="energoton")

        planner = Planner(pool)
        plans = planner.build_plans([e])

        p = Plan(
            [
                {"task": t1, "amount": 5, "assignee": e, "dry": (t1.id, 5)},
                {"task": t4, "amount": 2, "assignee": e, "dry": (t4.id, 2)},
            ]
        )
        p.commit(0, ())

        self.assertEqual(plans, (p,))

    def test_pool_after_plan(self):
        pool1 = Pool()
        t1 = Task(5, id_="1")
        t2 = Task(3, id_="2")
        pool1.add(t1)
        pool1.add(t2)

        root_pool = Pool()
        t3 = Task(4, id_="3")
        root_pool.add(t3)
        root_pool.add(pool1)

        e = NonDeterministicEnergoton(8)

        planner = Planner(root_pool)
        plans = planner.build_plans([e])

        result_pool = planner.pool_after_plan(plans[0])
        self.assertTrue(result_pool.children[t3.id].is_solved)
        self.assertFalse(result_pool.children[pool1.id].is_solved)

        self.assertEqual(
            result_pool.children[pool1.id].children[t1.id].spent, 1
        )

    def test_pool_two_energotons(self):
        pool1 = Pool()
        t1 = Task(5, id_="1")
        t2 = Task(3, id_="2")
        pool1.add(t1)
        pool1.add(t2)

        root_pool = Pool()
        t3 = Task(4, id_="3")
        root_pool.add(t3)
        root_pool.add(pool1)

        planner = Planner(root_pool)
        plans = planner.build_plans(
            [
                DeterministicEnergoton(8, id_="1"),
                DeterministicEnergoton(8, id_="2"),
            ]
        )

        for plan, result in zip(
            plans,
            (
                [("2", "1", 5), ("1", "2", 3), ("1", "3", 4)],
                [("1", "1", 5), ("1", "2", 3), ("2", "3", 4)],
            ),
        ):
            work_done = []
            for work in plan:
                work_done.append(
                    (work["assignee"].id, work["task"].id, work["amount"])
                )

            self.assertEqual(work_done, result)

    def test_several_cycles(self):
        pool = Pool()

        t1 = Task(5, id_="1")
        t2 = Task(2, id_="2")
        t3 = Task(4, id_="3")
        t4 = Task(2, id_="4")

        pool.add(t1)
        pool.add(t2)
        pool.add(t3)
        pool.add(t4)

        e = DeterministicEnergoton(8)

        planner = Planner(pool)
        plans = planner.build_plans([e], cycles=2)

        p = Plan(
            [
                {
                    "task": t1,
                    "amount": 5,
                    "assignee": e,
                    "cycle": 2,
                    "dry": (t1.id, 5),
                },
                {
                    "task": t2,
                    "amount": 2,
                    "assignee": e,
                    "cycle": 1,
                    "dry": (t2.id, 2),
                },
                {
                    "task": t3,
                    "amount": 4,
                    "assignee": e,
                    "cycle": 1,
                    "dry": (t3.id, 4),
                },
                {
                    "task": t4,
                    "amount": 2,
                    "assignee": e,
                    "cycle": 1,
                    "dry": (t4.id, 2),
                },
            ]
        )
        p.commit(0, ())

        self.assertEqual(plans, (p,))

    def test_by_cycles(self):
        pool = Pool()

        t1 = Task(5, id_="1")
        t2 = Task(2, id_="2")
        t3 = Task(4, id_="3")

        pool.add(t1)
        pool.add(t2)
        pool.add(t3)

        e = DeterministicEnergoton(8)

        planner = Planner(pool)
        plans = planner.build_plans([e], cycles=2)

        by_cycles = planner.by_cycles(plans)

        self.assertEqual(
            by_cycles,
            [
                {
                    2: [
                        {
                            "task": t1,
                            "amount": 5,
                            "assignee": e,
                            "cycle": 2,
                        },
                    ],
                    1: [
                        {
                            "task": t2,
                            "amount": 2,
                            "assignee": e,
                            "cycle": 1,
                        },
                        {
                            "task": t3,
                            "amount": 4,
                            "assignee": e,
                            "cycle": 1,
                        },
                    ],
                },
            ],
        )

    def test_by_assignees(self):
        pool = Pool()

        t1 = Task(2, id_="1")
        t2 = Task(4, id_="2")
        t3 = Task(2, id_="3")

        pool.add(t1)
        pool.add(t2)
        pool.add(t3)

        e1 = DeterministicEnergoton(4, id_="1")
        e2 = DeterministicEnergoton(4, id_="2")

        planner = Planner(pool)
        plans = planner.build_plans([e1, e2])

        by_assignees = planner.by_assignees(plans)

        self.assertEqual(
            by_assignees,
            [
                {
                    "1": [
                        {
                            "task": t1,
                            "amount": 2,
                            "assignee": e1,
                            "cycle": 1,
                        },
                        {
                            "task": t3,
                            "amount": 2,
                            "assignee": e1,
                            "cycle": 1,
                        },
                    ],
                    "2": [
                        {
                            "task": t2,
                            "amount": 4,
                            "assignee": e2,
                            "cycle": 1,
                        }
                    ],
                },
            ],
        )

    def test_build_plans_high_priority_blocked(self):
        pool = Pool()
        t1 = Task(4, id_="1", priority=Priority("high"))
        t2 = Task(5, id_="2", priority=Priority("low"))
        t3 = Task(6, id_="3", priority=Priority("highest"))

        pool.add(t1)
        pool.add(t2)
        pool.add(t3)

        Blocking(t2, t3)

        e = DeterministicEnergoton(8, id_="1")
        planner = Planner(pool)

        plans = planner.build_plans([e], 2)

        p1 = Plan(
            [
                {"task": t2, "amount": 5, "assignee": e, "cycle": 1},
                {"task": t3, "amount": 6, "assignee": e, "cycle": 2},
            ]
        )

        p1.commit(0, ())

        self.assertEqual(plans, (p1,))

    def test_build_plans_complex_priorities(self):
        t1 = Task(2, id_="1", priority=Priority("normal"))
        t2 = Task(2, id_="2", priority=Priority("normal"))
        t3 = Task(2, id_="3", priority=Priority("normal"))
        t4 = Task(2, id_="4", priority=Priority("normal"))

        t5 = Task(8, id_="5", priority=Priority("highest"))
        t6 = Task(8, id_="6", priority=Priority("highest"))

        t7 = Task(8, id_="7", priority=Priority("normal"))

        pool = Pool(children=[t1, t2, t3, t4, t5, t6, t7])

        Blocking(t5, t6)

        e = DeterministicEnergoton(8, id_="1")
        planner = Planner(pool)

        plans = planner.build_plans([e], 2)

        p1 = Plan(
            [
                {"task": t1, "amount": 2, "assignee": e, "cycle": 1},
                {"task": t2, "amount": 2, "assignee": e, "cycle": 1},
                {"task": t3, "amount": 2, "assignee": e, "cycle": 1},
                {"task": t4, "amount": 2, "assignee": e, "cycle": 1},
                {"task": t5, "amount": 8, "assignee": e, "cycle": 2},
            ]
        )
        p1.commit(0, ())

        p2 = Plan(
            [
                {"task": t5, "amount": 8, "assignee": e, "cycle": 1},
                {"task": t6, "amount": 8, "assignee": e, "cycle": 2},
            ]
        )
        p2.commit(0, ())

        self.assertEqual(plans, (p1, p2))

    def test_must_priority(self):
        pool = Pool()
        t1 = Task(4, id_="1", priority=Priority("must"))
        t2 = Task(4, id_="2", priority=Priority("low"))
        t3 = Task(4, id_="3", priority=Priority("highest"))

        pool.add(t1)
        pool.add(t2)
        pool.add(t3)

        Blocking(t2, t3)

        e = DeterministicEnergoton(4)
        planner = Planner(pool)

        plans = planner.build_plans([e])

        p1 = Plan(
            [
                {"task": t1, "amount": 4, "assignee": e, "cycle": 1},
            ]
        )

        p1.commit(0, ())

        self.assertEqual(plans, (p1,))



import unittest
from unittest import mock

from energoton import Pool, Task


class TestPool(unittest.TestCase):
    def test_task_by_key(self):
        task_id = "there-is-some-id"

        task = Task(10, id_=task_id, name="Task 1")
        pool = Pool(children=[task], name="Pool 1")

        self.assertEqual(pool.get(task_id), task)

        task_id2 = "there-is-another-id"
        task = Task(20, id_=task_id2, name="Task 2")
        pool.add(task)

        self.assertEqual(pool.get(task_id2), task)

    def test_get_embedded_task(self):
        task_id = "there-is-some-id"

        task = Task(10, id_=task_id, name="Task 1")
        pool = Pool(children=[task], name="Pool 1")

        root_pool = Pool(children=[pool], name="Root Pool")
        found_task = root_pool.get(task.id)
        self.assertEqual(found_task, task)

    def test_task_already_exists(self):
        same_id = "same-id"

        task = Task(10, id_=same_id, name="Task 1")
        pool = Pool(children=[task], name="Pool 1")

        task2 = Task(20, id_=same_id, name="Task 2")

        with self.assertRaises(ValueError):
            pool.add(task2)

    def test_parenting(self):
        task_id = "there-is-some-id"

        task = Task(10, id_=task_id, name="Task 1")
        pool = Pool(children=[task], name="Pool 1")

        self.assertEqual(task.parent, pool)

        pool.pop(task_id)

        self.assertIsNone(task.parent)

        new_pool = Pool(name="Pool 2")

        pool.add(new_pool)
        self.assertEqual(new_pool.parent, pool)

        pool.pop(new_pool.id)
        self.assertIsNone(new_pool.parent)

    def test_pop_embedded_task(self):
        task = Task(10, id_="1", name="Task 1")
        pool = Pool(children=[task], name="Pool 1")

        root_pool = Pool(children=[pool], name="Root Pool")
        found_task = root_pool.pop(task.id)
        self.assertEqual(found_task, task)

        self.assertEqual(len(root_pool), 1)
        self.assertEqual(len(pool), 0)

    def test_pop_from_embedded_task(self):
        task = Task(10, id_="1")
        p1 = Pool(children=[task], id_="lowest")
        p2 = Pool(children=[p1], id_="middle")
        p3 = Pool(children=[p2], id_="high")

        root_pool = Pool(children=[p3], id_="root")
        found_task = p3.pop(task.id)
        self.assertEqual(found_task, task)

        self.assertTrue(task.id not in root_pool.children)
        self.assertTrue(task.id not in p3.children)
        self.assertTrue(task.id not in p2.children)
        self.assertTrue(task.id not in p1.children)

    def test_iter(self):
        tasks = [Task(i, id_=i) for i in range(5)]
        pool = Pool(children=tasks, name="Pool 1")

        for task in pool:
            self.assertIn(task, tasks)

        self.assertEqual(len(pool), len(tasks))

    def test_is_solved(self):
        tasks = [Task(i, id_=i) for i in range(5)]
        pool = Pool(children=tasks, name="Pool 1")

        self.assertFalse(pool.is_solved)

        for task in tasks:
            task.work_done.append(
                {"task": task, "amount": task.cost, "assignee": mock.Mock()}
            )

        self.assertTrue(pool.is_solved)

    def test_done(self):
        task1 = Task(10, name="Task 1")
        task1.work_done.append(
            {"task": task1, "amount": task1.cost, "assignee": mock.Mock()}
        )

        task2 = Task(20, name="Task 2")

        pool = Pool(children=[task1, task2], name="Pool 1")

        self.assertEqual(list(pool.done), [task1])
        self.assertEqual(list(pool.todo), [task2])

    def test_composite(self):
        task1 = Task(10, name="Task 1")
        task2 = Task(20, name="Task 2")

        task2.work_done.append(
            {"task": task2, "amount": task2.cost, "assignee": mock.Mock()}
        )
        pool1 = Pool(children=[task1, task2], name="Pool 1")

        task3 = Task(30, name="Task 3")
        task4 = Task(40, name="Task 4")
        pool2 = Pool(children=[task3, task4], name="Pool 2")

        task5 = Task(50, name="Task 5")
        task6 = Task(60, name="Task 6")
        task6.work_done.append(
            {"task": task6, "amount": task6.cost, "assignee": mock.Mock()}
        )

        root_pool = Pool(
            children=[pool1, pool2, task5, task6], name="Root Pool"
        )

        self.assertFalse(root_pool.is_solved)

        self.assertEqual(list(root_pool.todo), [pool1, pool2, task5])
        self.assertEqual(list(root_pool.done), [task6])

        # make the first pool solved
        task1.work_done.append(
            {"task": task1, "amount": task1.cost, "assignee": mock.Mock()}
        )
        self.assertEqual(list(root_pool.todo), [pool2, task5])

        # make the second pool solved
        task3.work_done.append(
            {"task": task3, "amount": task3.cost, "assignee": mock.Mock()}
        )
        task4.work_done.append(
            {"task": task4, "amount": task4.cost, "assignee": mock.Mock()}
        )
        self.assertEqual(list(root_pool.todo), [task5])

        # make the last task solved
        task5.work_done.append(
            {"task": task5, "amount": task5.cost, "assignee": mock.Mock()}
        )
        self.assertTrue(root_pool.is_solved)

        self.assertEqual(list(root_pool.done), [pool1, pool2, task5, task6])
        self.assertEqual(list(root_pool.todo), [])

    def test_composite_iter(self):
        task1 = Task(10, name="Task 1")
        task2 = Task(20, name="Task 2")

        pool1 = Pool(children=[task1, task2], name="Pool 1")

        task3 = Task(30, name="Task 3")

        root_pool = Pool(children=[pool1, task3], name="Root Pool")
        self.assertEqual(list(root_pool), [pool1, task3])



import unittest
from unittest import mock

from energoton import Task, Priority
from energoton.planner import Plan


class TestPlan(unittest.TestCase):
    def test_value(self):
        tasks = (
            Task(5, priority=Priority("lowest")),
            Task(5, priority=Priority("low")),
            Task(5, priority=Priority("normal")),
            Task(5, priority=Priority("high")),
            Task(5, priority=Priority("highest")),
        )

        work_done = [
            {"task": t, "amount": 5, "assignee": mock.Mock(), "dry": (t.id, 5)}
            for t in tasks
        ]
        plan = Plan(work_done)
        plan.commit(31, ())

        self.assertEqual(plan.value, 31)



from distutils.core import setup

from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="energoton",
    version="0.2",
    description="Automated task planning package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="IlyaFaer",
    author_email="ilya.faer.gurov@gmail.com",
    url="https://github.com/IlyaFaer/energoton_py/",
    packages=[
        "energoton",
        "energoton.base",
        "energoton.energoton",
        "energoton.planner",
        "energoton.relation",
        "energoton.work",
    ],
)



"""Relations between work units."""

from energoton.base import Id


class Alternative(Id):
    """Alternative relation.

    Describes tasks, which can be replaced by each other.
    If one of the alternative tasks is solved, all the others
    are no longer considered by energotons and will not be solved.

    Args:
        alternatives (List[Task | Pool]):
            List of alternative tasks.
        id_ (Optional[str]):
            Id of the relation. Generated, if not provided.
    """

    def __init__(self, *alternatives, id_=None):
        super().__init__(id_)

        if len(alternatives) <= 1:
            raise ValueError(
                "Alternative relation must have at least two alternatives."
            )

        self.alternatives = alternatives

        for unit in alternatives:
            unit.relations[self.id] = self

    @property
    def is_solved(self):
        """Check if the alternative relation is solved.

        Returns:
            bool: True if one of the alternatives is solved, False otherwise.
        """
        return any(unit.is_solved for unit in self.alternatives)


class Blocking(Id):
    """Blocking relation.

    The blocked tasks is not considered by energotons until the
    blocker is solved.

    Args:
        blocker (Task | Pool):
        blocked (Task | Pool):
        id_ (Optional[str]):
            Id of the relation. Generated, if not provided.
    """

    def __init__(self, blocker, blocked, id_=None):
        super().__init__(id_)

        self.blocker = blocker
        self.blocked = blocked

        self.blocker.priority = (
            self.blocked.priority
            if self.blocked.priority.value > self.blocker.priority.value
            else self.blocker.priority
        )

        blocked.relations[self.id] = self
        blocker.relations[self.id] = self

    def drop(self):
        """Drop the blocking relation."""
        del self.blocked.relations[self.id]
        del self.blocker.relations[self.id]



"""Basic functionality classes."""

import uuid


class Id:
    """
    Object id functionality. Used by energotons,
    pools, relations and tasks.

    Args:
        id_ (Optional[Union[str, uuid.UUID]]):
            The object id. If not provided,
            a UUID is generated.
    """

    def __init__(self, id_=None):
        self.id = id_ or uuid.uuid4()

    def __eq__(self, other):
        """Equality operator.

        Args:
            other (Any): Object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return self.id == other.id



"""Planning related classes."""

import copy


class Plan(list):
    """A single work plan.

    Represents a possible work plan. Consists of dicts, each of which
    represents a piece of work done by some energoton on some task.
    """

    def commit(self, value, dry_pool):
        """Commit the plan and finalize its statistics.

        Args:
            value (int): The value of the plan.
            dry_pool (list): The dry pool after the plan.
        """
        self.value = value
        self.energy_spent = 0
        dry = []

        self.dry_pool_after_plan = {t["id"]: t for t in dry_pool}

        for w in self:
            self.energy_spent += w["amount"]
            dry.append((w["task"].id, w["amount"]))

        self.dry = tuple(dry)

    def __eq__(self, other):
        """Equality operator.

        Returns:
            bool: True if the plans are equal, False otherwise.
        """
        return self.dry == other.dry


class Planner:
    """
    Planner builds plans for the given pool
    and provides methods to analyze them.

    NOTE: On init, Planner makes a deep copy
    of the given pool. It sticks the planner
    to the pool, and allows to re-use the pool
    in multiple planners (e.g. you have 2 teams and
    you want to compare how they'll handle the same
    pool of tasks, running planners in parallel).

    Args:
        pool (work.pool.Pool):
            The pool of tasks to be solved.
    """

    def __init__(self, pool):
        if len(pool) == 0:
            raise ValueError(
                f"The given pool {pool.id} - {pool.name} is empty."
            )

        self._pool = copy.deepcopy(pool)
        self._dry_pool = self._pool.dry

        self._plans = [Plan()]

    def build_plans(self, energotons, cycles=1):
        """Build plans for the given energotons and amount of cycles.

        Energotons are copied on the method call, so they can be
        safely reused in multiple planners.

        Args:
            energotons (List[Energoton]):
                List of energotons to be used for planning.
            cycles (int):
                Number of work cycles to be performed.

        Return:
            tuple[Plan]: The best plans found.
        """
        if len(energotons) == 0:
            raise ValueError("No energotons provided for planning.")

        if cycles < 1:
            raise ValueError(
                "The number of work cycles must be greater than 0."
            )

        energotons = copy.deepcopy(energotons)
        for e in energotons:
            e.pool = self._pool

        self._plans = (Plan(),)

        for c in range(1, cycles + 1):
            for e in energotons:
                new_plans = []
                for plan in self._plans:
                    for new_plan in e.build_plans(
                        (
                            plan.dry_pool_after_plan
                            if getattr(plan, "dry_pool_after_plan", None)
                            else self.dry_pool_after_plan(plan)
                        ),
                        c,
                        plan,
                    ):
                        if new_plans:
                            if new_plan.value < new_plans[0].value:
                                continue

                            if new_plan.value > new_plans[0].value:
                                new_plans.clear()

                        if new_plan not in new_plans:
                            new_plans.append(new_plan)

                self._plans = tuple(new_plans)
                e.recharge()

        return self._plans

    def pool_after_plan(self, plan):
        """Return the pool state after the given plan.

        Args:
            plan (Plan): The plan to be applied.

        Returns:
            Pool: The pool state after the plan is executed.
        """
        pool = copy.deepcopy(self._pool)

        for work_done in plan:
            pool.children[work_done["task"].id].work_done.append(work_done)

        return pool

    def dry_pool_after_plan(self, plan):
        """Return the dry pool state after the given plan.

        Used for plan calculations.

        Args:
            plan (Plan): The plan to be applied.

        Returns:
            Dict[str, Dict[str, Any]]:
                The dry pool state after the plan is executed.
        """
        pool = copy.deepcopy(self._dry_pool)

        for work_done in plan:
            task = pool[work_done["task"].id]
            task["left"] -= work_done["amount"]
            if task["left"] == 0:
                del pool[work_done["task"].id]

        return pool

    @staticmethod
    def by_cycles(plans):
        """Group plans by cycles.

        Args:
            plans (List[Plan]): The plans to be grouped.

        Returns:
            List[Dict[str, List[Dict[str, Any]]]]:
                The plans grouped by cycles.
        """
        by_cycles = []
        for plan in plans:
            new_plan = {}
            for work_done in plan:
                if work_done["cycle"] not in new_plan:
                    new_plan[work_done["cycle"]] = []

                new_plan[work_done["cycle"]].append(work_done)

            by_cycles.append(new_plan)

        return by_cycles

    @staticmethod
    def by_assignees(plans):
        """Group plans by assignees.

        Args:
            plans (List[Plan]): The plans to be grouped.

        Returns:
            List[Dict[str, List[Dict[str, Any]]]]:
                The plans grouped by assignees.
        """
        by_assignees = []
        for plan in plans:
            new_plan = {}
            for work_done in plan:
                as_id = work_done["assignee"].id

                if as_id not in new_plan:
                    new_plan[as_id] = []

                new_plan[as_id].append(work_done)

            by_assignees.append(new_plan)

        return by_assignees



# flake8: noqa
from energoton.energoton import (
    DeterministicEnergoton,
    NonDeterministicEnergoton,
)
from energoton.planner import Planner
from energoton.relation import Alternative, Blocking
from energoton.work import Task, Pool, Priority
